/*
 * ADC-LF.c
 */ 

#define F_CPU 3333333
#define BAUD_9600 4*F_CPU/9600
char uart_str[64];

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

/*
In this exercise we will set up and use the ADC (analog-digital converter). We will output the voltage over UART.
*/ 

#define TX_PORT		PORTB
#define TX_PIN		0
#define KLAKK       0x00


void uart_init(unsigned long baud){
	//Sets up registers for UART
	
	TX_PORT.OUTSET = (1 << TX_PIN); 	//Setting up TX pin as output
	TX_PORT.DIRSET = (1 << TX_PIN); 	//Setting up TX pin as output
	
	//Set baud rate register
	USART3.BAUDL = (uint8_t) baud;			//Set baud rate without shifting to get the 8 low bits
	USART3.BAUDH = (uint8_t)(baud >> 8);	//Shift register right by 8 bits to get the 8 high bits
	
	
	//USART.CTRLC CMODE bits default to async, 1 stop bit, 8 bit character size
	//Since all bits are default 0 we only need to change the character size part
	USART3.CTRLC = (0x3 << USART_CHSIZE0_bp);
	
	//Enable RX and TX
	USART3.CTRLB = (1 << USART_RXEN_bp) | (1 << USART_TXEN_bp);

	//Enable interrupts on incoming data
	USART3.CTRLA |= (1 << USART_RXCIE_bp); 
}

void uart_transmit(char data){	
	//Sends a character over UART
	
	//First we should check that there isn't already data being sent
	//	if there is, we should probably wait for it to finish first
	while (!(USART3.STATUS & (1 << USART_DREIF_bp))){
		//wait for previous transmit to finish
	};
	
	//Put our new data into tx data register
	USART3.TXDATAL = data;
}


//To send a string we can do this
void uart_transmit_string(char* data) {
	while (*data != '\0') {
		uart_transmit(*data);
		data++;
	}
}

//helper function for converting data from adc into a nice readable format, and sending it over UART
void uart_print_voltage(uint16_t data) {
    uint16_t data2 = data / 4;  //yes
    uint8_t voltage_integer = data2*323ul/100000ul; //This is a lot faster than floats. Compiler magic
	uint8_t voltage_fraction = (data2*323ul%100000ul)/1000;

	sprintf(uart_str, "Voltage: %u.%02u V\n",voltage_integer,voltage_fraction); //format string and put voltage  
    uart_transmit_string(uart_str);
}

void adc_init() {
	ADC0.CTRLA |= (ADC_RESSEL_10BIT_gc); //Set resolution, we choose 10 bits
	
	ADC0.CTRLB |= (ADC_SAMPNUM_ACC4_gc); //OPTIONAL: We can use multiple samples if we like, example here with 4
		//More samples gives a better result but takes longer.
	
	ADC0.CTRLC |= (ADC_REFSEL_VDDREF_gc); //We select to use the supply voltage (VDD) as voltage reference
	ADC0.CTRLC |= (ADC_PRESC_DIV2_gc); //ADC clock prescaler, best accuracy when run below 2MHz. (Here div2 ~1.46 MHz)
	ADC0.CTRLA |= (ADC_ENABLE_bm); //Enable the ADC
}

uint16_t adc_read(uint8_t channel){
	ADC0.MUXPOS = channel; //Select input on the ADC mux
	//NOTE: We can use = here because this is the only thing the register holds. Neat!
	
	ADC0.INTFLAGS |= (ADC_RESRDY_bm); //Clear the results ready flag
	
	ADC0.COMMAND |= (ADC_STCONV_bm); //Start a conversion
	
	while(!(ADC0.INTFLAGS & ADC_RESRDY_bm)){
		 //Wait for the results ready flag to be set
	}
	return ADC0.RES; //Return 10 bit result 

}

int main(void)
{
	//Initialize the UART with our function. 
	//We will be using a baudrate of 9600 (defined as BAUD_9600 at the top of the file)
	uart_init(BAUD_9600);
    adc_init();
	sei(); //Enable inerrupt, important for anything here to work
	
    uint16_t data;
    while (1) 
    {
		//get voltage from adc_read and send it over UART
        data=adc_read(KLAKK); //update data each loop using polling.
        uart_print_voltage(data); //print voltage over UART
    }
}


//Interrupt service routine for the UART receiver. Doesn't actually do anything with the received data.
ISR (USART3_RXC_vect) {
	//In the interrupt we will read the data in the receive buffer
	
	//First we should check that new data has arrived the receive buffer
	while (!(USART3.STATUS & (1 << USART_RXCIF_bp))){
		//wait for previous transmit to finish
	};
	
	//Store the data in a temporarily variable
	uint8_t tmp = USART3.RXDATAL; 
}